/* This is a sample application for the LM32 running on the TS-4300 with a
 * TS-8100 baseboard.
 */

#include "ts4300.h"

/* This test function demonstrates that the compiler correctly handles tail
 * recursion.  Otherwise this program would suffer a stack overflow.
 */
void recursion_test(int x) {
	POKE32((SHM + x) & ~0x3, x);
	if (x) recursion_test(x - 1);
}

int main(void)
{
	unsigned int i, prev, cur, diff, t, head, tail;
	signed int x;

	ack_startup();
	outstr("\nTS-8100 LM32 demo application.\n\n");

	// read baseboard ADC:
	POKE16(ADC_CONF, 0x18);
	POKE16(ADC_MASK, 0x3f);
	usleep(400000);
	for(i=1; i<=6; i++) {
		x = (signed short)PEEK16(ADC_DAT(i));
		if (i > 2) x = (x * 1006)/200; // account for resistor dividers
		x = (x * 2048)/0x8000; // scale
		outstr("ADC channel ");
		outint(i);
		outstr(": ");
		outint(x);
		outstr("mV.\n");
	}

	/* This basic benchmark is testing how fast a simple loop iterates.
	 * The result on the current FPGA codebase shows that each iteration
	 * takes 21 clock cycles.  This is reasonable because no opcode can 
	 * go faster than 3 cycles.  At a minimum, this loop needs 5 or 6
	 * opcodes.  I'm not sure if the 21 result is 7 opcodes * 3 cycles, or
	 * if the store to data memory is slower due to bus arbitration.
	 */
	prev=PEEK32(TIMER);
	recursion_test(0x1fff);
	diff=PEEK32(TIMER) - prev;
	outstr("8k memory writes using recursion took ");
	outint(diff);
	outstr("us.\n");
	for (i=0; i!=0x2000; i+=4) if (PEEK32(SHM + i) != i) {
		outstr("Recursion test FAILED. i=");
		outint(i);
		outstr("\n");
		shutdown();
	}
	prev=PEEK32(TIMER);
	for (i=0x1fff; i; i--) POKE32((SHM + i) & ~0x3, i);
	diff=PEEK32(TIMER) - prev;
	outstr("8k memory writes using iteration took ");
	outint(diff);
	outstr("us.\n");

	// baseboard access:
	POKE32(MUXBUS_CONFIG, 0x321);
	outstr("Baseboard ID: 0x");
	outhex(PEEK16(MUXBUS16));
	outstr("\n\nInitial DIO state: ");
	prev = PEEK16(MUXBUS16 + 0xc);
	outhex(prev);
	outstr("\n\nStarting DIO monitoring:\n\n");

	/* For DIO monitoring, this code does two writes to RAM every time it
	 * sees a DIO change.  The first write is the timestamp and the 2nd is
	 * the new DIO state.  The scratch region from 0x12000 to 0x13fff is
	 * used as a circular buffer.
	 *
	 * In a real application, the LM32 would store the data and the ARM
	 * host would process it.  This way the DIO monitoring would maintain
	 * guaranteed real time performance.  With this sample code, it stops
	 * to do user IO every 4096 microseconds.  The IO can be time
	 * consuming, so this creates a window where DIO glitches may not be
	 * detected.
	 */
	POKE32(SHM + 4, prev);
	head = tail = 8;
	while (1) {
		t = PEEK32(TIMER);
		cur = PEEK16(MUXBUS16 + 0xc);
		if (prev != cur) {
			POKE32(SHM + tail, t);
			POKE32(SHM + 4 + tail, cur);
			tail += 8;
			if (tail == 0x2000) tail = 0;
			prev = cur;
		}

		if ((0 == (t & 0xfff)) && (head != tail)) { // process data
			cur = PEEK32(SHM + ((head + 0x1ffc) & 0x1fff));
			diff = PEEK32(SHM + 4 + head) ^ cur;
			i = 0;
			while (diff) {
				if (diff & 1) {
					if (cur & 1) outstr("-");
					else outstr("+");
					outint(i);
					outstr(" ");
					outhex(PEEK32(SHM + head));
					outstr("\n");
				}
				cur >>= 1;
				diff >>= 1;
				i++;
			}
			head += 8;
			if (head == 0x2000) head = 0;
			
		}
	}

	shutdown();
}

