/* This is a sample application for the LM32 running on the TS-4300 with a
 * TS-8500 baseboard.
 */

#include "ts4300.h"

/* This test function demonstrates that the compiler correctly handles tail
 * recursion.  Otherwise this program would suffer a stack overflow.
 */
void recursion_test(int x) {
	POKE32((SHM + x) & ~0x3, x);
	if (x) recursion_test(x - 1);
}

int main(void)
{
	unsigned int cur, i, diff, t, head, tail;
	signed int x;
	unsigned int prev[4];

	ack_startup();
	outstr("\nTS-8500 LM32 demo application.\n\n");

	/* This basic benchmark is testing how fast a simple loop iterates.
	 * The result on the current FPGA codebase shows that each iteration
	 * takes 21 clock cycles.  This is reasonable because no opcode can 
	 * go faster than 3 cycles.  At a minimum, this loop needs 5 or 6
	 * opcodes.  I'm not sure if the 21 result is 7 opcodes * 3 cycles, or
	 * if the store to data memory is slower due to bus arbitration.
	 */
	prev[0]=PEEK32(TIMER);
	recursion_test(0x1fff);
	diff=PEEK32(TIMER) - prev[0];
	outstr("8k memory writes using recursion took ");
	outint(diff);
	outstr("us.\n");
	for (i=0; i!=0x2000; i+=4) if (PEEK32(SHM + i) != i) {
		outstr("Recursion test FAILED. i=");
		outint(i);
		outstr("\n");
		shutdown();
	}
	prev[0]=PEEK32(TIMER);
	for (i=0x1fff; i; i--) POKE32((SHM + i) & ~0x3, i);
	diff=PEEK32(TIMER) - prev[0];
	outstr("8k memory writes using iteration took ");
	outint(diff);
	outstr("us.\n");

	outstr("\n\nInitial DIO state: ");
	for (i=0; i<4; i++) {
		prev[i] = PEEK32(0x60 + 4*i);
		outstr("\n");
		outhex(prev[i]);
	}
	outstr("\n\nStarting DIO monitoring:\n\n");

	/* For DIO monitoring, this code does four writes to RAM every time it
	 * sees a DIO change.  The first write is the timestamp.  The next is
	 * the ID of the DIO bank, 0 to 3.  The others are the before and after
	 * contents of that DIO bank.  The entire scratch region from 0x12000
	 * to 0x13fff is used as a circular buffer.
	 *
	 * In a real application, the LM32 would store the data and the ARM
	 * host would process it.  This way the DIO monitoring would maintain
	 * guaranteed real time performance.  With this sample code, it stops
	 * to do user IO every 4096 microseconds.  The IO can be time
	 * consuming, so this creates a window where DIO glitches may not be
	 * detected.
	 */
	head = tail = 0;
	while (1) {
		t = PEEK32(TIMER);
		for (i=0; i<4; i++) {
			cur = PEEK32(0x60 + 4*i);
			if (prev[i] != cur) {
			POKE32(SHM + tail, t);
			POKE32(SHM + tail + 4, i);
			POKE32(SHM + tail + 8, prev[i]);
			POKE32(SHM + tail + 12, cur);
			tail += 16;
			if (tail == 0x2000) tail = 0;
			prev[i] = cur;
			}
		}

		if ((0 == (t & 0xfff)) && (head != tail)) { // process data
			cur = PEEK32(SHM + head + 8);
			diff = PEEK32(SHM + head + 12) ^ cur;
			i = PEEK32(SHM + head + 4) << 5;
			while (diff) {
				if (diff & 1) {
					if (cur & 1) outstr("-");
					else outstr("+");
					outint(i);
					outstr(" ");
					outhex(PEEK32(SHM + head));
					outstr("\n");
				}
				cur >>= 1;
				diff >>= 1;
				i++;
			}
			head += 16;
			if (head == 0x2000) head = 0;
			
		}
	}

	shutdown();
}

